/*
 * Copyright (c) 2013-2016 Solarflare Communications Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the FreeBSD Project.
 */

#ifndef	_SYS_EFX_PHY_IDS_H
#define	_SYS_EFX_PHY_IDS_H

#define	EFX_PHY_NULL	0

typedef enum efx_phy_type_e {			/* GENERATED BY scripts/genfwdef */
	EFX_PHY_TXC43128 = 1,
	EFX_PHY_SFX7101 = 3,
	EFX_PHY_QT2022C2 = 4,
	EFX_PHY_PM8358 = 6,
	EFX_PHY_SFT9001A = 8,
	EFX_PHY_QT2025C = 9,
	EFX_PHY_SFT9001B = 10,
	EFX_PHY_QLX111V = 12,
	EFX_PHY_QT2025_KR = 17,
	EFX_PHY_AEL3020 = 18,
	EFX_PHY_XFI_FARMI = 19,
} efx_phy_type_t;


#endif	/* _SYS_EFX_PHY_IDS_H */
