/*-
 * This file is provided under a dual BSD/GPLv2 license. When using or
 * redistributing this file, you may do so under either license.
 *
 *   BSD LICENSE
 *
 * Copyright 2010-2012 Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * * Neither the name of the above-listed copyright holders nor the
 * names of any contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 *   GPL LICENSE SUMMARY
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __NETCFG_H
#define __NETCFG_H

#include <fman.h>
#include <argp.h>

/* Configuration information related to a specific ethernet port */
struct fm_eth_port_cfg {
	/**< A list of PCD FQ ranges, obtained from FMC configuration */
	struct list_head *list;
	/**< The "Rx default" FQID, obtained from FMC configuration */
	uint32_t rx_def;
	/**< Other interface details are in the fman driver interface */
	struct fman_if *fman_if;
};

struct netcfg_info {
	uint8_t num_ethports;
	/**< Number of ports */
	struct fm_eth_port_cfg port_cfg[0];
	/**< Variable structure array of size num_ethports */
};

struct interface_info {
	char *name;
	struct ether_addr mac_addr;
	struct ether_addr peer_mac;
	int mac_present;
	int fman_enabled_mac_interface;
};

struct netcfg_interface {
	uint8_t numof_netcfg_interface;
	uint8_t numof_fman_enabled_macless;
	struct interface_info interface_info[0];
};

/* pcd_file: FMC netpcd XML ("policy") file, that contains PCD information.
 * cfg_file: FMC config XML file
 * Returns the configuration information in newly allocated memory.
 */
struct netcfg_info *netcfg_acquire(void);

/* cfg_ptr: configuration information pointer.
 * Frees the resources allocated by the configuration layer.
 */
void netcfg_release(struct netcfg_info *cfg_ptr);

#ifdef RTE_LIBRTE_DPAA_DEBUG_DRIVER
/* cfg_ptr: configuration information pointer.
 * This function dumps configuration data to stdout.
 */
void dump_netcfg(struct netcfg_info *cfg_ptr);
#endif

#endif /* __NETCFG_H */
